--[[
===================================
 IWD (Improved World Defense) v1.6
===================================
Author: tattooedpierre (Duathorthad <The Infernals> Emerald Dream EU)

Desscription: 	IWD will intercept WorldDefense messages and display the information
				neatly in a frame just above your main chat window.
				
Credits:		Thanks to Torrin for the original mod, and also to the DefenseTracker addon, 
				whose code to track new zones I adapted for IWD.			
]]

--[[Global Variables]]--
IWD = {}
IWD_VERSION = "1.6";
IWD_DEBUG = false; 		-- debugging tool on?
IWD_HITS = {};		-- timestamp array for every hit seen in the last IWD_TFRAME seconds
IWD_SORT = {};		-- array of sorted places
IWD_COUNTER = {};	-- counter array used by interval
IWD_LASTHIT = "";	-- latest place under attack
IWD_LASTTIME = 6;	-- seconds since last hit
IWD_LASTSECOND = 0;	-- used to prevent redundant code in onupdates
IWD_ALERT = "igPVPUpdate";

--[[Fubar/LDB Variables]]--
IWD_HEAVY = false;		-- FuBar: for icon flashing; tells if there is heavy attack
IWD_LDBHIT = {};		-- FuBar: store 5 most recent attacks
IWD_COUNT = 1;			-- FuBar: counts attacks
IWD_SIMPLE = "|cffffffff-none-";	-- FuBar: name of the attacked area

--[[Configuration variables]]--
IWDConfiguration = {		
	ENABLE = true,			-- enable IWD
	MUTE = false,			-- mute WorldDefense spam
	SOUND = false,			-- audio alerts
	SUBZONES = true, 		-- show sub-zones
	TFRAME = 300,			-- time in seconds to drop hits
	THRESHOLD = 1,			-- the number of hits in IWD_TFRAME before this place is displayed
	INTERVAL = 5,			-- chat window notification interval
	LOCK = true, 			-- lock the alert frame
	LDBONLY = false
}
--------------
-- Hooks
--------------
local Pre_ChatFrame_AddMessageEventFilter;

--[[Start LDB Code]]--

local pName = "IWD_LDB"
local f = CreateFrame("frame")
local ldb = LibStub:GetLibrary("LibDataBroker-1.1")

local colors = {
	["mail_new"] = "|cff00ff00",
	["mail_old"] = "|ccccccccc",
}
local dataobj = ldb:GetDataObjectByName(pName) or ldb:NewDataObject(pName, {
	type = "data source",
	text = "IWD",
	icon = "Interface\\Icons\\Ability_Rogue_Shadowstrikes",
})

--[[ End LDB Code]]--

------------------
-- OnLoad Event
------------------
function IWD.OnLoad()
	local i;
	SlashCmdList["IWD"] = IWDCmdLine;
	SLASH_IWD1 = "/iwd";

	this:RegisterEvent("CHAT_MSG_CHANNEL");
	this:RegisterEvent("ZONE_CHANGED_NEW_AREA");
	this:RegisterEvent("ZONE_CHANGED");
	this:RegisterEvent("ZONE_CHANGED_INDOORS");
	
	if (not IWD_newSubzones) then 
		IWD_newSubzones = {}; 
	end
	
	IWD_cleanNewSubzones();
	
	-- Hook chat messages function for muting
	if (IWDConfiguration.ENABLE) then
		Pre_ChatFrame_AddMessageEventFilter = ChatFrame_AddMessageEventFilter;
		ChatFrame_AddMessageEventFilter = IWD_ChatFrame_AddMessageEventFilter;
	end

	if (DEFAULT_CHAT_FRAME) then
		DEFAULT_CHAT_FRAME:AddMessage("|cffedc300Improved World Defense v"..IWD_VERSION.."|cffffffff AddOn loaded.  |c00ff00ff/iwd|cffffffff for options menu.");
	end

end
--------------------
-- Parse Chat Event
--------------------
function IWD.OnEvent(event)
	-- arg1: "<place> is under attack!"
	-- arg4: "3. WorldDefense"
	-- arg7: 23
	-- arg8: 3
	-- arg9: "WorldDefense"

	if (event == "CHAT_MSG_CHANNEL" and arg9 == "WorldDefense" and arg2 == "") then
		IWD.Msg("IWD Debug", arg1);
		IWDAddHit(arg1);
	end
	
	if ((event == "ZONE_CHANGED_NEW_AREA")
		or (event == "ZONE_CHANGED")
		or (event == "ZONE_CHANGED_INDOORS")) then
		IWD.HandleZoneChange();
	end
end

function IWD_ChatFrame_AddMessageEventFilter(event)

	-- let all messages through except WorldDefense under attack messages
	if (event ~= "CHAT_MSG_CHANNEL" or arg9 ~= "WorldDefense" or arg2 ~= "") then
		Pre_ChatFrame_AddMessageEventFilter(event);
	end
end

function IWDAddHit(arg1)
	local place, k, i;
	local timenow = math.floor(GetTime());
	local tmparr;
	IWD_HEAVY = false;
	_, _, place = string.find(arg1, "^|c........(.+) is under attack!|r$");
	if (place) then
		-- first hit?
		if not (IWD_HITS[place]) then
			IWD_HITS[place] = {};
		end

		-- 2nd sorted array to reduce burden on OnUpdate
		i = 1;
		tmparr = IWD_SORT;	-- numeric indexed array required for any kind of sorting
		IWD_SORT = {};
		for _, k in pairs(tmparr) do
			-- include every place seen hit recently except newest
			if (k ~= place) then
				IWD_SORT[i] = k;
			end
			i = i + 1;
		end
		-- put most recent hit last.  since this is run every hit, every one ends up sorted
		IWD_SORT[i] = place;

		-- add hit w/ timestamp in HITS array
		table.insert(IWD_HITS[place], timenow);
		IWD_LASTHIT = place;
		IWD_LASTTIME = 0;

		-- Increment counter
		if not(IWD_COUNTER[place]) then
			IWD_COUNTER[place] = 0;
		end
		IWD_COUNTER[place] = IWD_COUNTER[place] + 1;
		-- if not muted, and this place has been hit enough to notify the chat window, and place hits meet threshold
		if (IWDConfiguration.MUTE == false and IWD_COUNTER[place] >= IWDConfiguration.INTERVAL and IWDConfiguration.THRESHOLD <= table.getn(IWD_HITS[place])) then
		if (IWDConfiguration.SOUND) then -- if audio alerts are active
			PlaySound(IWD_ALERT);
		end
			if (IWDConfiguration.LDBONLY == false) then
				IWD.Msg("IWD","|cffff0000"..place.." is under heavy attack!");
			end
			IWD_COUNTER[place] = 0;
			IWD_HEAVY = true;
		end

		-- force an update so we don't have to wait until the next second to display new hit
		IWD_LASTSECOND = 0;
	else
	--	IWD.Msg("IWD Error","No WorldDefense pattern match!");
	end
end

function IWDFormatTime(timenow, timethen)
	local d, h, m, s;
	local time = timenow - timethen;
	local text;

	if (time <= 0) then
		return "00:00";
	elseif (time < 3600) then
		d, h, m, s = ChatFrame_TimeBreakDown(time);
		return format("%02d:%02d", m, s);
	else
		return "1 hr+";
	end
end
------------------
-- IWD OnUpdate
------------------
function IWD.OnUpdate()
	local v,k,i;
	local msgColor;
	local timenow = math.floor(GetTime());	-- remove fractions of a second

	-- only execute once per second
	if (IWDConfiguration.ENABLE and IWD_LASTSECOND < timenow) then
		IWD_LASTSECOND = timenow;

		local i, k, v, n;
		local text = "";
		-- iterate through each location seen attacked recently
		for n, k in pairs(IWD_SORT) do				-- n = numerical order, k = place
			-- iterate through timestamps
			v = table.getn(IWD_HITS[k]);		-- v = number of hits at place k
			for i = 1, v do				-- i = individual hit gettime
				-- if more than IWF_TFRAME seconds old, remove from array
				if (IWD_HITS[k][i] and timenow - IWD_HITS[k][i] > IWDConfiguration.TFRAME) then
					table.remove(IWD_HITS[k], i);
					v = v - 1;
					if (v == 0) then
						-- last hit removed, remove place entirely
						IWD_HITS[k] = nil;
						IWD_COUNTER[k] = nil;
						-- delete SORT entry
						IWD_SORT[n] = nil;
					end
				end
			end

			-- if there are any recent hits (the last hit may have gotten removed above)
			-- and if the number of hits >= the current threshold
			if (v > 0 and v >= IWDConfiguration.THRESHOLD) then
				-- then add place and time since last hit to frontstring

				-- if the name of the subzone under attack is in our table
				-- then add it to the text output
	
				if (v >= math.ceil(IWDConfiguration.INTERVAL/2)) then -- set FuBar colouring
					msgColor = "|cffedc000";
				else
					msgColor = "|cffffffff";
				end
				if (v >= IWDConfiguration.INTERVAL) then
					msgColor = "|cffff0000";
				end
												
				if (IWD_COUNT >= 9) then
					IWD_COUNT = 1;
				end	

				 -- if we want to show zones + sub-zones AND they are different AND not unknown				
				if ((IWDConfiguration.SUBZONES) and not(k == IWD_ZoneList[k]) and not(IWD_ZoneList[k] == nil)) then
					IWD.Msg("IWD Debug","Show subZones (!=): "..v.."//"..k);
					text = text.."["..v.."] "..k.." in "..IWD_ZoneList[k].." ("..IWDFormatTime(timenow, IWD_HITS[k][v])..")\n";
					IWD_SIMPLE = msgColor..k.." in "..IWD_ZoneList[k].." ["..v.."]"; -- LDB Text
					
				-- we don't want subzones anyway, don't have different names or havent encountered this zone before..
				elseif ((k == IWD_ZoneList[k]) or (IWDConfiguration.SUBZONES == false) or (IWD_ZoneList[k] == nil)) then 					
					IWD.Msg("IWD Debug","Show subZones (=): "..v.."//"..k);
					text = text.."["..v.."] "..k.." ("..IWDFormatTime(timenow, IWD_HITS[k][v])..")\n";					
					IWD_SIMPLE = msgColor..k.." ["..v.."]"; -- LDB Text
				end
			end
		end
		
		if (IWDConfiguration.LDBONLY == false) then
			IWDRecentHits:SetText(text);
		end
		dataobj.text = IWD_SIMPLE; -- Set LDB Text
		
		-- last place hit color change
		if (IWD_LASTTIME <= 6) and (IWDConfiguration.LDBONLY == false) then
			if (IWD_LASTTIME <= 5) then
				-- reduce the brightness of red every second for 5 seconds
				IWDLastHit:SetFont("Fonts\\skurri.ttf", 14);
				IWDLastHit:SetTextColor(1 - (IWD_LASTTIME * .2), 0, 0);
				IWDLastHit:SetText(IWD_LASTHIT);
			else
				-- erase last place hit if longer than 5 seconds
				IWDLastHit:SetText("");
			end

			IWD_LASTTIME = IWD_LASTTIME + 1;
		end
	end
end

function IWD.Msg(reason, text) -- beautify messages a bit
	if ((IWD_DEBUG == false) and (reason == "IWD Debug")) then
		-- do nothing
	else
		DEFAULT_CHAT_FRAME:AddMessage("|cffedc300"..reason..": |cffffffff"..text, 1, 1, 0);
	end
end

-----------------
-- IWD Options
-----------------
function IWDEnable()
	if (IWDConfiguration.ENABLE == false) then
		IWD.Msg("IWD","Addon Enabled.");
		IWDConfiguration.ENABLE = true;
		Pre_ChatFrame_AddMessageEventFilter = ChatFrame_AddMessageEventFilter;
		ChatFrame_AddMessageEventFilter = IWD_ChatFrame_AddMessageEventFilter;
		IWDFrame:RegisterEvent("CHAT_MSG_CHANNEL");
	else
		IWD.Msg("IWD","Addon is already enabled.", 1, 1, 0);
	end
end

function IWDDisable()
	if (IWDConfiguration.ENABLE) then
		IWD.Msg("IWD","IWD Disabled.");
		IWDConfiguration.ENABLE = false;
		ChatFrame_AddMessageEventFilter = Pre_ChatFrame_AddMessageEventFilter;
		IWDFrame:UnregisterEvent("CHAT_MSG_CHANNEL");
	else
		IWD.Msg("IWD","Addon is already disabled.");
	end
end

function IWDCmdLine(param)
	local arg1;
	local arg2;
	_, _, arg1, arg2 = string.find(param, "(%a+)%s+(%w+)");
	if not(arg1) then
		arg1 = param;
	end

	if (arg1 == "enable") then
		IWDEnable();
	elseif (arg1 == "disable") then
		IWDDisable();
	elseif (arg1 == "tframe") then
		if (arg2) then
			arg2 = tonumber(arg2);
			if (arg2 and arg2 > 0) then
				IWD.Msg("IWD","Time frame set to "..arg2.." seconds");
				IWDConfiguration.TFRAME = arg2;
			end
		end
	elseif (arg1 == "mute") then
		IWD.Msg("IWD",IWDConfiguration.MUTE);
		if (IWDConfiguration.MUTE) then
			IWD.Msg("IWD","WorldDefense Unmuted.");
			IWDConfiguration.MUTE = false;
		else
			IWD.Msg("IWD","WorldDefense Muted.");
			IWDConfiguration.MUTE = true;
		end
	elseif (arg1 == "threshold") then
		if (arg2) then
			arg2 = tonumber(arg2);
			if (arg2) then
				IWD.Msg("IWD","Threshold set to "..arg2.." hits");
				IWDConfiguration.THRESHOLD = arg2;
			end
		end	
	elseif (arg1 == "inject") then
		if (arg2) then
			IWD.Msg("IWD Debug","Injecting attack to "..arg2.."..");
			IWDAddHit(arg2.." is under attack!");
		end
	elseif (arg1 == "interval") then
		if (arg2) then
			arg2 = tonumber(arg2);
			if (arg2) then
				DEFAULT_CHAT_FRAME:AddMessage("Interval set to "..arg2.." hits", 1, 1, 0);
				IWDConfiguration.INTERVAL = arg2;
			end
		end
	elseif (arg1 == "lock") then
		if (IWDConfiguration.LOCK) then
			IWD.Msg("IWD","Alert frame UNLOCKED");
			IWDFrame:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
                                  edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
                                  tile = true, tileSize = 16, edgeSize = 16, 
                                  insets = { left = 4, right = 4, top = 4, bottom = 4 }});
			IWDFrame:SetBackdropColor(0,0,0,1);
			IWDConfiguration.LOCK = false;
		else
			IWD.Msg("IWD","Alert frame LOCKED");
			IWDConfiguration.LOCK = true;
			IWDFrame:SetBackdrop(nil);
		end
	elseif (arg1 == "help" or arg1 == "?") then
		IWD.Msg("IWD","Improved World Defense v"..IWD_VERSION.." Slash Commands:", 1, 1, 0);
		IWD.Msg("IWD","enable -- Enable IWD");
		IWD.Msg("IWD","disable -- Disable IWD");
		IWD.Msg("IWD","mute -- Mutes chat alerts");
		IWD.Msg("IWD","tframe <number> -- Set the timeframe");
		IWD.Msg("IWD","threshold <number> -- Set threshold");
		IWD.Msg("IWD","interval <number> -- Set notification interval");
		IWD.Msg("IWD","lock -- Locks/Unlocks the notification window to move");
		if (IWDConfiguration.ENABLE) then
			IWD.Msg("IWD","Current Status: ENABLED");
		else
			IWD.Msg("IWD","Current Status: DISABLED");
		end
	else
		--IWDOptions:Show();
		InterfaceOptionsFrame_OpenToCategory("IWD")
	end
end

------------------------------
--  Zone Tracking Functions --
------------------------------
function IWD.HandleZoneChange()
	local subzone = GetSubZoneText();
	local zone = GetZoneText();
	
	if (subzone == nil) then
		subzone = zone;
	end
	if (subzone == "") then
		subzone = zone;
	end

	IWD_updateZoneList(subzone, zone);	
end

function IWD_updateZoneList(subzone, zone)
	if (IWD_ZoneList[subzone] == nil) then
		if (IWD_newSubzones[subzone] == nil) then
			-- Subzone has not been seen before
			IWD_newSubzones[subzone] = zone;
			DEFAULT_CHAT_FRAME:AddMessage("|cffedc300IWD: |cffffffff(|cffffffff".. subzone.. "//".. zone.. "|cffffffff)|cffffffff |cffff0000New zone definition saved!");
			return 1;
		end
	end
	
	return 0;
end

function IWD_cleanNewSubzones()
	local IWD_oldNewSubzones = IWD_newSubzones;
	IWD_newSubzones = {};
	
	for loc in pairs(IWD_oldNewSubzones) do
		if (IWD_ZoneList[loc] == nil) then
			IWD_newSubzones[loc] = IWD_oldNewSubzones[loc];
		end
	end
end